from PyQt5.QtXml import QDomDocument
from PyQt5.QtGui import QFont, QPolygonF
from qgis.core import QgsComposition, QgsComposerLabel, QgsProject, \
    QgsComposerPolygon, QgsComposerPolyline, QgsLineSymbol
from PAGGF.db import get_connection
import os.path
from PyQt5.QtCore import QPointF


class PagesGenerator(object):
    """
    Generates pages for the CID PV (i.e. pages of adjacents, pages of
    commission members, ...)
    """
    def __init__(self, pv_gen):
        self.pv_gen = pv_gen

        self.pos_title_x = 20  # X-Position of the title of a page
        self.pos_title_y = 20  # Y-position of the title of a page
        self.pos_title_separator_x1 = 20  # Position of the line over the title
        self.pos_title_separator_x2 = 190
        self.pos_title_separator_y = 15

        self.number_of_elements_per_page = None  # how many elements are
        # displayed on a page
        self.title = None  # title of the section

    def add_title(self, title_text):
        # generate separator

        # generate label
        self.add_composer_label(
            title_text, self.pos_title_x, self.pos_title_y,
            font=QFont("Liberation Sans", 10, QFont.Bold))

        polygon = QPolygonF()
        polygon.append(QPointF(20, 15))
        polygon.append(QPointF(190, 15))

        self.polyline = QgsComposerPolyline(polygon, self.pv_gen.pv)
        self.pv_gen.pv.addComposerPolyline(self.polyline)

        self.polyline.setItemPosition(
            self.polyline.pagePos().x(), self.polyline.pagePos().y(),
            page=self.pv_gen.current_page_number)

        style = QgsLineSymbol.createSimple(
            {'color': '0,0,0,255', 'width': '0.3', 'capstyle': 'square'})
        self.polyline.setPolylineStyleSymbol(style)

    def add_composer_label(self, text, x, y, font=None):
        """Adds a composer label in the PV"""
        label = QgsComposerLabel(self.pv_gen.pv)
        label.setText(text)

        if font:
            label.setFont(font)

        label.adjustSizeToText()
        self.pv_gen.pv.addComposerLabel(label)
        label.setItemPosition(x, y, page=self.pv_gen.current_page_number)

    def add_composer_label_for_el(
            self, i, text, offset_x=0, offset_y=0, font=None):
        """
        Adds a composer label for an element in the PV.

        Each element is in a cell. There is a maximal number of element
        in each page.
        """
        self.add_composer_label(
            text,
            self.cell_origin_x(i) + offset_x,
            self.cell_origin_y(i) + offset_y,
            font=font)

    def generate(self, parcel):
        elements = self.get_elements(parcel)

        # on crée une nouvelle page avec un title
        self.pv_gen.new_page()
        self.add_title('{} ({})'.format(self.title, len(elements)))

        page_with_title = True
        for i, el in enumerate(elements):
            if i % self.number_of_elements_per_page == 0:
                if page_with_title:  # si c'est la page avec le titre -> pas
                    # de nouvelle page car fait avant le titre
                    page_with_title = False
                else:  # sinon nouvelle page
                    self.pv_gen.new_page()

            self.add_element_in_pv(i, el)

    def get_elements(self, parcel):
        """Return the elements to display on the PV"""
        pass  # A implémenter

    def add_element_in_pv(self, i, el):
        """Add a single element into the pv"""
        pass  # A implémenter

    def cell_origin_y(self, i):
        """Top position for the cell i on a page"""
        pass  # A implémenter

    def cell_origin_x(self, i):
        """Left position for the cell i on a page"""
        pass  # A implémenter


class AdjacentsPageGenerator(PagesGenerator):
    def __init__(self, pv_gen):
        super(AdjacentsPageGenerator, self).__init__(pv_gen)
        self.number_of_elements_per_page = 17
        self.title = 'Liste des limitrophes'

    def cell_origin_y(self, i):
        return 30 + (15 * (i % self.number_of_elements_per_page))

    def cell_origin_x(self, i):
        return 20

    def get_elements(self, parcel):
        """Return the adjacents that are stored in the db for an inventory"""
        conn = get_connection()
        cur = conn.cursor()
        cur.execute(
            "\
SELECT firstname, lastname FROM idc.inventories_adjacent_people \
WHERE id_inventory = {};".format(parcel['id_inventory'])
            )

        return cur.fetchall()

    def add_element_in_pv(self, i, el):
        name = el[0] # TODO plus a jour firstname + lastname

        self.add_composer_label_for_el(i, 'Nom / prénom:')
        self.add_composer_label_for_el(
            i, name, offset_x=30,
            font=QFont("Liberation Sans", 10, QFont.Bold))
        self.add_composer_label_for_el(i, 'Signature :', offset_x=100)
        self.add_composer_label_for_el(
            i, 'Si refus, motif de refus :', offset_y=5)


class CommissionMemberPageGenerator(PagesGenerator):
    def __init__(self, pv_gen):
        super(CommissionMemberPageGenerator, self).__init__(pv_gen)
        self.number_of_elements_per_page = 14
        self.title = 'Liste des membres CID'

    def cell_origin_y(self, i):
        return 30 + (20 * (i % self.number_of_elements_per_page))

    def cell_origin_x(self, i):
        return 20

    def get_elements(self, parcel): #TODO seulement la requete sql
        """
        Return the member of the commission (except witness) that are
        stored in the db for an inventory
        """
        conn = get_connection()
        cur = conn.cursor()
        cur.execute("\
SELECT cm.firstname, cm.lastname, cm_role.name FROM idc.commission_members AS cm \
INNER JOIN idc.commission_member_enum AS cm_role \
    ON(cm_role.id = cm.member_quality_id) \
WHERE id_inventory = {} \
    AND cm_role.name != 'Témoin';".format(parcel['id_inventory'])
                    )

        return cur.fetchall()

    def add_element_in_pv(self, i, el):
        name = el[0]  # TODO plus a jour firstname + lastname
        quality = el[2]

        self.add_composer_label_for_el(i, 'Qualité')
        self.add_composer_label_for_el(
            i, quality, offset_x=30)
        self.add_composer_label_for_el(i, 'Nom / prénom:', offset_y=5)
        self.add_composer_label_for_el(
            i, name, offset_x=30, offset_y=5,
            font=QFont("Liberation Sans", 10, QFont.Bold))
        self.add_composer_label_for_el(
            i, 'Signature :', offset_x=100, offset_y=5)
        self.add_composer_label_for_el(
            i, 'Si refus, motif de refus :', offset_y=10)


class WitnessPageGenerator(PagesGenerator):
    def __init__(self, pv_gen):
        super(WitnessPageGenerator, self).__init__(pv_gen)
        self.number_of_elements_per_page = 17
        self.title = 'Liste des personnes témoin'

    def cell_origin_y(self, i):
        return 30 + (15 * (i % self.number_of_elements_per_page))

    def cell_origin_x(self, i):
        return 20

    def get_elements(self, parcel): #TODO modif
        """Return the witness that are stored in the db for an inventory"""
        conn = get_connection()
        cur = conn.cursor()
        cur.execute("\
SELECT cm.firstname, cm.lastname, cm_role.name FROM idc.commission_members AS cm \
INNER JOIN idc.commission_member_enum AS cm_role \
    ON(cm_role.id = cm.member_quality_id) \
WHERE id_inventory = {} \
    AND cm_role.name = 'Témoin';".format(parcel['id_inventory'])
                    )

        return cur.fetchall()

    def add_element_in_pv(self, i, el):
        name = el[0] # TODO plus a jour firstname + lastname

        self.add_composer_label_for_el(i, 'Nom / prénom:')
        self.add_composer_label_for_el(
            i, name, offset_x=30,
            font=QFont("Liberation Sans", 10, QFont.Bold))
        self.add_composer_label_for_el(
            i, 'Signature :', offset_x=100)
        self.add_composer_label_for_el(
            i, 'Si refus, motif de refus :', offset_y=5)


class CIDPvGenerator(object):
    """Generate pdf file that contains the CID PV"""
    def __init__(self, plugin_dir, parcel):
        self.pv = None  # The pv that is generated
        self.number_of_pages_before_adjacents = 4  # number of page before
        # printing the adjacents
        self.current_page_number = 1  # the number of the page where we
        # are (starts at 1)

        self.parcel = parcel
        self.plugin_dir = plugin_dir

    def new_page(self):
        self.current_page_number += 1
        self.pv.setNumPages(self.current_page_number)

    def create_pdf_at(self, out_path):
        """Generate the pdf at a given path"""
        template_path = os.path.join(
            self.plugin_dir, 'composer/cid_inventory.qpt')

        template_content = None
        with open(template_path, encoding="utf-8") as template_file:
            template_content = template_file.read()
        template_file.close()

        template_qdoc = QDomDocument()
        template_qdoc.setContent(template_content)

        self.pv = QgsComposition(QgsProject.instance())
        self.pv.loadFromTemplate(
            template_qdoc,
            substitutionMap={'QGIS_PLUGIN_DIR': self.plugin_dir})

        self.current_page_number = self.pv.numPages()

        adjacents_gen = AdjacentsPageGenerator(self)
        adjacents_gen.generate(self.parcel)

        com_member_gen = CommissionMemberPageGenerator(self)
        com_member_gen.generate(self.parcel)

        witness_gen = WitnessPageGenerator(self)
        witness_gen.generate(self.parcel)

        self.pv.setAtlasMode(QgsComposition.PreviewAtlas)
        self.pv.atlasComposition().prepareForFeature(self.parcel)
        self.pv.exportAsPDF(out_path)
