from PyQt5.QtCore import QUrl
from PyQt5.QtGui import QIcon, QDesktopServices
from PyQt5.QtWidgets import QAction, QLabel  # QFormLayout, QDialog
from qgis.core import QgsMessageLog, QgsVectorLayer, QgsProject, QgsGeometry,\
    QgsFeature, QgsApplication
from qgis.gui import QgsMapToolIdentifyFeature, QgsMapToolEmitPoint, QgsGui
from .db import get_connection
# Initialize Qt resources from file resources.py
from .resources import *
from .paggf_layers import LoadedLayers
import os.path
import tempfile
import base64
from . import paggf_version
from .ui.CID_load_attr_from_shp import LoadAttrFromFile


class PAGGFCore(object):
    """Encodes the CID inventory"""

    def __init__(self, iface, version, menu_title):
        """
        Constructor

        Keyword arguments:
        iface -- the QgsInterface
        version -- version of the plugin (BM, CID or EF see paggf_version.py)
        menu_title -- the tile of the menu
        """
        self.iface = iface
        self.version = version
        self.menu_title = menu_title

        self.mapCanvas = iface.mapCanvas()

        self.paggf_actions_configuration = dict()  # configuration of the
        # PAGGF actions

        self.load_parcel_dlg = None
        self.load_attributes_dlg = None

        self.toolbar = self.iface.addToolBar(u'CID')
        self.toolbar.setObjectName(u'CID')
        self.layers = LoadedLayers()

        self.maptool = None

        self.actions = dict()  # Dictionary where are stored the actions
        self.active_action_name = None  # The name of the action which is
        # active at the  moment
        self.stop_action_function = dict()  # Dictionary of function to
        # execute at the end of an action

        self.plugin_dir = os.path.dirname(__file__)

    def enable_action_when_feature_selected(
            self, paggf_schema_layer, paggf_table_layer, action_name):
        """
        Enable an action when one and only one feature is selected in a given
        layer
        """
        layer = self.layers.get(paggf_schema_layer, paggf_table_layer)

        def get_set_slot_checkable(layer, action_name):
            def ret():
                if layer.selectedFeatureCount() == 1:
                    self.actions[action_name].setEnabled(True)
                else:
                    self.actions[action_name].setEnabled(False)
            return ret

        get_set_slot_checkable(layer, action_name)()
        layer.selectionChanged.connect(
            get_set_slot_checkable(layer, action_name))

    def enable_paggf_action(self, action_id, configuration):
        """
        Enable a paggf action. To be called during __init__.

        Keyword arguments:
        action_id -- The id of the action
        configuration -- its configuration
        """
        self.paggf_actions_configuration[action_id] = configuration

    def create_paggf_menu(self):
        """
        Create the PAGGF menu (the tree of the menu is defined in the
        varialbe self.paggf_menu_content)
        """
        self.paggf_menu = self.iface.mainWindow().menuBar().addMenu(
            self.menu_title)
        self.paggf_menu.setObjectName('mPAGGF')

        def def_recursive_create_menu(parent, children):
            """
            Ajout recursif d'un menu
            parent  -> deja un menu
            la suite (à ajouter) soit
                - une liste (sous-menu)
                - une action (fin du la branche du menu)
                - un label (action pas encore ecrite)
            """
            for child in children:
                if type(child) == QAction:  # fin de l'arble -> ajoute l'action
                    parent.addAction(child)
                elif type(child) == tuple:  # ajout sous menu
                    c_label, c_children = child
                    c_menu = parent.addMenu(c_label)
                    def_recursive_create_menu(c_menu, c_children)
                elif type(child) == str:  # Ajout d'une action non ecrite
                    action = parent.addAction(child)
                    action.setEnabled(False)
                else:  # Error
                    print(child)
                    print(type(child))

        def_recursive_create_menu(self.paggf_menu, self.paggf_menu_content())

    def create_action(
        self, name, icon_path, text, callback, enabled_flag=True,
        add_to_toolbar=True, status_tip=None,
        whats_this=None, parent=None, checkable=False
            ):
        """Create a toolbar icon to the toolbar.

        :param name: Name to get back the action from the dictionary of
            actions. The name will be the key of this dictionary.
        :type name: str

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :param checkable: Flag indicating if the action is checkable.
            Defaults to False.
        :type checkable: bool

        :returns: The action that was created. Note that the action is also
            added to self.actions dict.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.toolbar.addAction(action)

        if checkable:
            action.setCheckable(True)

        self.actions[name] = action
        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        if 'load_parcel_from_file' in self.paggf_actions_configuration:
            config = self.paggf_actions_configuration['load_parcel_from_file']
            if self.version == paggf_version.BM:
                callback = self.load_parcel_from_file
            else:
                callback = self.get_load_parcel_callback(config)

            self.create_action(
                'load_parcel_from_file',
                ':/plugins/PAGGF/img/icon_add_parcel.svg',
                text=config['menu_title'],
                callback=callback,
                parent=self.iface.mainWindow())  # , checkable=True)

        if 'load_parcel_manually' in self.paggf_actions_configuration:
            config = self.paggf_actions_configuration['load_parcel_manually']
            self.create_action(
                'load_parcel_manually',
                ':/plugins/PAGGF/img/icon_capture_parcel.svg',
                text=config['menu_title'],
                callback=self.get_load_parcel_callback(config),
                parent=self.iface.mainWindow())  # , checkable=True)

        if 'add_vertex' in self.paggf_actions_configuration:
            self.create_action(
                'add_vertex', ':/plugins/PAGGF/img/icon_add_vertex.png',
                text='Ajouter un sommet', callback=self.add_vertex,
                parent=self.iface.mainWindow(), checkable=True)
            self.stop_action_function['add_vertex'] = self.commit_changes

        if 'move_vertex' in self.paggf_actions_configuration:
            self.create_action(
                'move_vertex', ':/plugins/PAGGF/img/icon_move_vertex.png',
                text='Déplacer les sommets', callback=self.move_vertex,
                parent=self.iface.mainWindow(), checkable=True)
            self.stop_action_function['move_vertex'] = self.commit_changes

        if 'delete_vertex' in self.paggf_actions_configuration:
            self.create_action(
                'delete_vertex', ':/plugins/PAGGF/img/icon_delete_vertex.png',
                text='Supprimer des sommets', callback=self.delete_vertex,
                parent=self.iface.mainWindow(), checkable=True)
            self.stop_action_function['delete_vertex'] = self.commit_changes

        if 'add_ring' in self.paggf_actions_configuration:
            self.create_action(
                'add_ring', ':/plugins/PAGGF/img/icon_add_ring.svg',
                text='Ajouter un anneau', callback=self.add_ring,
                parent=self.iface.mainWindow(), checkable=True)
            self.stop_action_function['add_ring'] = self.commit_changes

        if 'reshape' in self.paggf_actions_configuration:
            self.create_action(
                'reshape', ':/plugins/PAGGF/img/icon_reshape.svg',
                text='Remodeler les entités', callback=self.reshape,
                parent=self.iface.mainWindow(), checkable=True)
            self.stop_action_function['reshape'] = self.commit_changes

        if 'edit_data' in self.paggf_actions_configuration:
            config = self.paggf_actions_configuration['edit_data']
            self.create_action(
                'edit_data',
                ':/plugins/PAGGF/img/icon_form_view.svg',
                text=config['menu_title'],
                callback=self.edit_data,
                parent=self.iface.mainWindow(), checkable=True)
            self.stop_action_function['edit_data'] = self.commit_changes

        if 'print_pv' in self.paggf_actions_configuration:  # TODO:NE PLUS UTILISER
            config = self.paggf_actions_configuration['print_pv']
            self.create_action(
                'print_pv',
                ':/plugins/PAGGF/img/icon_print_pv.png',
                text=config['menu_title'],
                callback=self.print_pv,
                parent=self.iface.mainWindow(), checkable=True)
            self.stop_action_function['print_pv'] = self.pan_trigger

        if 'archive_parcel' in self.paggf_actions_configuration:
            self.create_action(
                'archive_parcel',
                ':/plugins/PAGGF/img/icon_archive.svg',
                text='Archiver une parcelle', callback=self.archive_parcel,
                parent=self.iface.mainWindow(), checkable=True)
            self.stop_action_function['archive_parcel'] = self.commit_changes

        if 'copy_parcel' in self.paggf_actions_configuration:
            self.create_action(
                'copy_parcel',
                ':/plugins/PAGGF/img/icon_archive_copy.svg',
                text='Copier une parcelle archivée', callback=self.copy_parcel,
                parent=self.iface.mainWindow(), checkable=True)
            self.stop_action_function['copy_parcel'] = self.commit_changes
            
        if 'pv_upload_signatures' in self.paggf_actions_configuration:
            config = self.paggf_actions_configuration['pv_upload_signatures']
            self.create_action(
                'pv_upload_signatures',
                ':/plugins/PAGGF/img/icon_sign.svg',
                text=config['menu_title'],
                callback=self.pv_upload_signatures,
                parent=self.iface.mainWindow(), checkable=True)

            self.enable_action_when_feature_selected(
                config['pg_schema_name'],
                config['pg_table_name'],
                'pv_upload_signatures')

        if 'pv_w_signatures_print' in self.paggf_actions_configuration:
            config = self.paggf_actions_configuration['pv_w_signatures_print']
            self.create_action(
                'pv_w_signatures_print',
                ':/plugins/PAGGF/img/icon_print_pv2.png',
                text=config['menu_title'],
                callback=self.pv_w_signatures_print,
                parent=self.iface.mainWindow(), checkable=True)

            self.enable_action_when_feature_selected(
                config['pg_schema_name'],
                config['pg_table_name'],
                'pv_w_signatures_print')

        if 'load_attributes_from_file' in self.paggf_actions_configuration:
            config = self.paggf_actions_configuration[
                'load_attributes_from_file']
            self.create_action(
                'load_attributes_from_file',
                ':/plugins/PAGGF/img/icon_upload.png',
                text=config['menu_title'],
                callback=self.loadattributes_callback,
                parent=self.iface.mainWindow(), checkable=True)

        self.create_paggf_menu()

        dlg = self.welcome_dlg_class()
        dlg.exec_()

    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions.values():
            self.iface.removeToolBarIcon(action)
        # remove the toolbar
        del self.toolbar

        # remove the menu
        self.paggf_menu.clear()
        self.paggf_menu.deleteLater()

    def get_load_parcel_callback(self, config):
        def ret():
            dlg_class = config['dlg_class']
            widget_class = config['load_widget_class']

            if (not self.load_parcel_dlg) or \
                    (not self.load_parcel_dlg.isVisible()):
                self.load_parcel_dlg = dlg_class(
                    data_loader_class=widget_class,
                    mapCanvas = self.mapCanvas)
                self.load_parcel_dlg.show()
                self.load_parcel_dlg.exec_()
        return ret

    def load_parcel_from_file(self):  # TODO renommer
        """Displays a dialog with a form getting information about :
            - the hill
            - the subhill
            - the id of the parcel
            - the shapefile containing the vertex
        """

        if self.version == paggf_version.CID:
            pass
            # ne doit pas arriver
        else:
            # TODO changer
            config = self.paggf_actions_configuration['load_parcel_from_file']

            if (not self.load_parcel_dlg) or \
                    (not self.load_parcel_dlg.isVisible()):
                self.load_parcel_dlg = config['dlg_class']()
                self.load_parcel_dlg.cancel_ok_btns.accepted.connect(
                    self.load_and_display_shp)
                self.load_parcel_dlg.cancel_ok_btns.rejected.connect(
                    self.load_parcel_dlg.close)
                self.load_parcel_dlg.show()
                self.load_parcel_dlg.exec_()

    def load_and_display_shp(self):
        self.load_parcel_dlg.close()

        def create_vector_layer():
            layer_file_path = self.load_parcel_dlg.shp_path()
            layer_file_path_ext = os.path.splitext(layer_file_path)[1]

            provider_lib = "ogr"  # Default provider lib  for shp
            path = self.load_parcel_dlg.shp_path() # Default uri for shp
            if layer_file_path_ext.lower() == '.csv':
                provider_lib = "delimitedtext"
                path = "\
file://{}?delimiter={}&crs=epsg:32735&xField={}&yField={}".format(
                    self.load_parcel_dlg.shp_path(),
                    ";", "X", "Y")
            elif layer_file_path_ext.lower() == '.gpx':
                provider_lib = "gpx"
                path = "{}?type=waypoint".format(
                    self.load_parcel_dlg.shp_path())

            layer = QgsVectorLayer(
                path, "data_import", provider_lib)

            if layer.isValid():
                return layer
            else:
                return None

        if self.version == paggf_version.BM:
            # First check if the person has chose via the
            # commission_reference_choice field
            id_parcel = self.load_parcel_dlg.\
                commission_reference_choice.currentData()

            if not id_parcel:
                id_parcel = self.load_parcel_dlg.\
                    idc_parcel_code_choice.currentData()

            if id_parcel:
                conn = get_connection()
                cur = conn.cursor()

                layer = create_vector_layer()

                cur.execute("delete from gis.vertices_to_import")
                values = []
                for f in layer.getFeatures():
                    wkt = 'srid=32735;' + f.geometry().exportToWkt()
                    values.append("('{}','{}'::geometry)".format(
                        f['name'], wkt))

                query = "insert into gis.vertices_to_import (name, geom) \
values "
                query += ','.join(values)
                cur.execute(query)

                cur.execute("select gis.import_vertices(%s)", (id_parcel,))

                conn.commit()

                l = self.layers.get("gis", "parcels_to_valid")
                # update extent of affected layers
                if l:
                    l.updateExtents()
                    vertices_to_valid = self.layers.get(
                        "gis", "vertices_to_valid")
                    if vertices_to_valid:
                        vertices_to_valid.updateExtents()

                    isolated_vertices = self.layers.get(
                        "gis", "isolated_vertices")
                    if isolated_vertices:
                        isolated_vertices.updateExtents()

                    segments_parcels_to_valid = self.layers.get(
                        "gis", "segments_parcels_to_valid")
                    if segments_parcels_to_valid:
                        segments_parcels_to_valid.updateExtents()
                    # zoom to the new feature
                    self.mapCanvas.zoomToFeatureIds(l, [id_parcel])

            else:
                iface.messageBar.pushWarning(
                    'Erreur', 'pas de IDC selectionné')

        else:
            pass
            # CID GERER DANS CLASS

    def enable_disable_action_vertex(self, action_name):
        """
            Activate / desactivate the icon actions
            Returns a flat to know if the action is enable of disable
        """
        action = self.actions[action_name]
        new_active_action = True

        if self.active_action_name:
            active_action = self.actions[self.active_action_name]
            active_action.setChecked(False)

            if self.active_action_name in self.stop_action_function:
                # there exists a method to execute at the end of the action
                # so we have to execute this method
                self.stop_action_function[self.active_action_name]()

            if action_name == self.active_action_name:
                # no more action is active
                new_active_action = False
                self.active_action_name = None

        if new_active_action:
            self.active_action_name = action_name
            action.setChecked(True)

        return new_active_action

    def add_vertex(self):
        enable_action = self.enable_disable_action_vertex('add_vertex')

        if enable_action:
            add_vertex_layer = self.layers.get(
                'gis', 'segments_idc_parcels_to_valid')
            self.iface.setActiveLayer(add_vertex_layer)
            self.maptool = QgsMapToolEmitPoint(self.mapCanvas)
            self.maptool.canvasClicked.connect(self.add_vertex_at)
            self.mapCanvas.setMapTool(self.maptool)

    def add_vertex_at(self, new_vertex_point):
        add_vertex_layer = self.layers.get(
            'gis', 'segments_idc_parcels_to_valid')
        add_vertex_layer.startEditing()

        new_vertex_geom = QgsGeometry.fromPoint(new_vertex_point)
        new_vertex_buffer = new_vertex_geom.buffer(
            (self.mapCanvas.mapUnitsPerPixel() * 2), 0)
        new_vertex_rect = new_vertex_buffer.boundingBox()

        selected_edges = add_vertex_layer.getFeatures(new_vertex_rect)

        edge = QgsFeature()
        if selected_edges.nextFeature(edge):
            pt_wkt_txt = "'srid=32735;point({} {})'::geometry"\
                .format(new_vertex_point.x(), new_vertex_point.y())
            sql = 'SELECT gis.add_idc_vertex_to_parcel({},{},{},{})'\
                .format(
                    edge['id_parcel'], edge['id_vertex_before'],
                    edge['id_vertex_after'], pt_wkt_txt)
            transaction = add_vertex_layer.dataProvider().transaction()
            transaction.executeSql(sql)

            # FIXME est-ce que ca peut ê simplifié? (copier/coller action Hugo)
            for l in QgsProject.instance().mapLayers().values():
                if '"gis"."idc_vertices_to_valid"' in l.source():
                    # make_layer_dirty(l, 'id')
                    # def make_layer_dirty(layer, field_name):
                    layer = l
                    field_name = 'id'
                    feature = next(layer.getFeatures())
                    fake_value = feature.attribute(field_name)
                    fake_field_id = feature.fieldNameIndex(field_name)
                    # sets id to id just to dirty buffer
                    layer.editBuffer().changeAttributeValue(
                        feature.id(), fake_field_id, fake_value)
                    l.triggerRepaint()
            add_vertex_layer.triggerRepaint()
        else:
            print('NO selection')
        add_vertex_layer.commitChanges()

    def move_vertex(self):
        enable_action = self.enable_disable_action_vertex('move_vertex')

        if enable_action:
            vertex_layer = self.layers.get('gis', 'idc_vertices_to_valid')
            self.iface.setActiveLayer(vertex_layer)
            vertex_layer.startEditing()
            self.iface.actionMoveFeature().trigger()

    def commit_changes(self):
        vertex_layer = self.layers.get('gis', 'idc_vertices_to_valid')
        if vertex_layer:  # TODO FIXME faut-il vraiment le faire ? 
            vertex_layer.commitChanges()
        self.pan_trigger()

    def delete_vertex(self):
        enable_action = self.enable_disable_action_vertex('delete_vertex')

        if enable_action:
            vertex_layer = self.layers.get('gis', 'idc_vertices_to_valid')
            self.iface.setActiveLayer(vertex_layer)

            self.maptool = QgsMapToolIdentifyFeature(
                self.mapCanvas, vertex_layer)
            self.maptool.featureIdentified.connect(
                self.delete_identified_vertex)
            self.mapCanvas.setMapTool(self.maptool)

    def delete_identified_vertex(self, feat):
        if(feat):
            vertex_layer = self.layers.get('gis', 'idc_vertices_to_valid')
            vertex_layer.startEditing()
            vertex_layer.deleteFeature(feat.id())
            vertex_layer.commitChanges()
            self.iface.mapCanvas().refresh()

    def add_ring(self, feat):
        """
        Trigger the addRing action on the layer idc_parcels_to_valid. This
        action allow the user to create ring into a polygon
        """
        enable_action = self.enable_disable_action_vertex('add_ring')

        if enable_action:
            parcels_layer = self.layers.get('gis', 'idc_parcels_to_valid')
            self.iface.setActiveLayer(parcels_layer)
            parcels_layer.startEditing()
            self.iface.actionAddRing().trigger()

    def reshape(self, feat):
        """
        Trigger the reshape / the cut feature action on the layer
        idc_parcels_to_valid. This action allow the user to cut external part
        of a polygone
        """
        enable_action = self.enable_disable_action_vertex('reshape')

        if enable_action:
            parcels_layer = self.layers.get('gis', 'idc_parcels_to_valid')
            self.iface.setActiveLayer(parcels_layer)
            parcels_layer.startEditing()
            reshapeAction = [
                o for o in self.iface.mainWindow().findChildren(QAction)
                if o.objectName() == 'mActionReshapeFeatures'][0]
            reshapeAction.trigger()

    def edit_data(self):
        enable_action = self.enable_disable_action_vertex('edit_data')
        if enable_action:
            config = self.paggf_actions_configuration['edit_data']
            parcels_layer = self.layers.get(
                config['pg_schema_name'], config['pg_table_name'])
            self.iface.setActiveLayer(parcels_layer)
            parcels_layer.startEditing()
            self.iface.actionIdentify().trigger()

    def print_pv(self):
        enable_action = self.enable_disable_action_vertex('print_pv')

        if enable_action:
            # FIXME tjs meme code a regrouper
            config = self.paggf_actions_configuration['print_pv']
            parcel_layer = self.layers.get(
                config['pg_schema_name'], config['pg_table_name'])
            self.iface.setActiveLayer(parcel_layer)
            self.maptool = QgsMapToolEmitPoint(self.mapCanvas)
            self.maptool.canvasClicked.connect(self.print_pv_for_parcel_at)
            self.mapCanvas.setMapTool(self.maptool)

    def print_pv_for_parcel_at(self, point_clicked):
        # FIXME idem tjs meme code
        config = self.paggf_actions_configuration['print_pv']
        parcel_layer = self.layers.get(
            config['pg_schema_name'], config['pg_table_name'])

        click_geom = QgsGeometry.fromPoint(point_clicked)
        click_buffer = click_geom.buffer(
            (self.mapCanvas.mapUnitsPerPixel() * 2), 0)
        click_rect = click_buffer.boundingBox()

        selected_parcels = parcel_layer.getFeatures(click_rect)

        parcel = QgsFeature()
        if selected_parcels.nextFeature(parcel):
            out_path = os.path.join(
                tempfile.gettempdir(),
                "inventory_{}.pdf".format(parcel['id']))

            pv_generator = config['pv_generator_cls'](self.plugin_dir, parcel)
            pv_generator.create_pdf_at(out_path)

            # ask the OS to open the generated PDF
            QDesktopServices.openUrl(QUrl.fromLocalFile(out_path))

    def archive_parcel(self):
        enable_action = self.enable_disable_action_vertex('archive_parcel')
        if enable_action:
            # FIXME tjs meme code a regrouper
            config = self.paggf_actions_configuration['archive_parcel']
            parcel_layer = self.layers.get(
                config['pg_schema_name'], config['pg_table_name'])
            self.iface.setActiveLayer(parcel_layer)
            self.maptool = QgsMapToolEmitPoint(self.mapCanvas)
            self.maptool.canvasClicked.connect(self.archive_parcel_at)
            self.mapCanvas.setMapTool(self.maptool)

    def archive_parcel_at(self, point_clicked):
        # FIXME idem tjs meme code
        config = self.paggf_actions_configuration['archive_parcel']
        parcel_layer = self.layers.get(
            config['pg_schema_name'], config['pg_table_name'])

        click_geom = QgsGeometry.fromPoint(point_clicked)
        click_buffer = click_geom.buffer(
            (self.mapCanvas.mapUnitsPerPixel() * 2), 0)
        click_rect = click_buffer.boundingBox()

        selected_parcels = parcel_layer.getFeatures(click_rect)

        parcel = QgsFeature()
        if selected_parcels.nextFeature(parcel):
            print(parcel, parcel.attributes())
            conn = get_connection()
            cur = conn.cursor()
            cur.execute("select validate_object('land.parcels', %s, 'Validation CID')", (parcel['id'],))
            conn.commit()
            self.iface.mapCanvas().refresh()

    def copy_parcel(self):
        enable_action = self.enable_disable_action_vertex('copy_parcel')
        if enable_action:
            # FIXME tjs meme code a regrouper
            config = self.paggf_actions_configuration['copy_parcel']
            parcel_layer = self.layers.get(
                config['pg_schema_name'], config['pg_table_name'])
            self.iface.setActiveLayer(parcel_layer)
            self.maptool = QgsMapToolEmitPoint(self.mapCanvas)
            self.maptool.canvasClicked.connect(self.copy_parcel_at)
            self.mapCanvas.setMapTool(self.maptool)

    def copy_parcel_at(self, point_clicked):
        # FIXME idem tjs meme code
        config = self.paggf_actions_configuration['copy_parcel']
        parcel_layer = self.layers.get(
            config['pg_schema_name'], config['pg_table_name'])

        click_geom = QgsGeometry.fromPoint(point_clicked)
        click_buffer = click_geom.buffer(
            (self.mapCanvas.mapUnitsPerPixel() * 2), 0)
        click_rect = click_buffer.boundingBox()

        selected_parcels = parcel_layer.getFeatures(click_rect)

        parcel = QgsFeature()
        if selected_parcels.nextFeature(parcel):
            print(parcel, parcel.attributes())
            conn = get_connection()
            cur = conn.cursor()
            cur.execute("select copy_object('land.parcels', %s)", (parcel['id'],))
            conn.commit()

    def pv_upload_signatures(self):
        from .ui.CID_merge_pv_and_signatures import CIDPvMergeDlg
        from .ui.file_custom_widget import FileCustomWidget

        config = self.paggf_actions_configuration['pv_upload_signatures']
        parcel_layer = self.layers.get(
            config['pg_schema_name'], config['pg_table_name'])

        if parcel_layer.selectedFeatureCount() == 1:
            parcel_layer.startEditing()
            feature = parcel_layer.selectedFeatures()[0]

            self.upload_doc_dlg = CIDPvMergeDlg()
            self.upload_doc_widget = FileCustomWidget(self.upload_doc_dlg)

            file_custom_widget_label = QLabel()
            file_custom_widget_label.setBuddy(self.upload_doc_widget)
            file_custom_widget_label.setText('Enregistrement des signatures')

            self.upload_doc_dlg.formLayout.addRow(
                file_custom_widget_label, self.upload_doc_widget)

            self.upload_doc_widget.setEnabled(True)
            self.upload_doc_widget.set_data_base64(
                feature['doc_idc_report_signatures'])

            def update_file_widget(layer, feature):
                # update custom file
                layer.changeAttributeValue(
                    feature.id(),
                    layer.fields().indexFromName("doc_idc_report_signatures"),
                    self.upload_doc_widget.data_base64())

            self.upload_doc_dlg.buttonBox.accepted.connect(
                lambda lyr=parcel_layer, f=feature: update_file_widget(lyr, f)
            )

            def download_empty_signatures(parcel):
                def ret():
                    generated_pv_filename = os.path.join(
                        tempfile.gettempdir(),
                        "inventory_pv_{}.pdf".format(parcel['id']))

                    pv_generator = config['pv_generator_cls'](
                        self.plugin_dir, parcel)
                    pv_generator.create_pdf_at(generated_pv_filename)

                    empty_signatures_filename = os.path.join(
                        tempfile.gettempdir(),
                        "inventory_{}_signatures.pdf".format(parcel['id']))

                    with open(generated_pv_filename, "rb") as generated_pv_file,\
                            open(empty_signatures_filename, 'wb') as empty_sign_file:
                        from .deps.PyPDF2__v1_26_0.PyPDF2 import PdfFileReader, PdfFileWriter
                        pv_reader = PdfFileReader(generated_pv_file)
                        writer = PdfFileWriter()
                        for page in pv_reader.pages[4:]:
                            writer.addPage(page)
                        writer.write(empty_sign_file)

                    QDesktopServices.openUrl(
                        QUrl.fromLocalFile(empty_signatures_filename))
                return ret

            self.upload_doc_dlg.download_empty_signatures_button.clicked \
                .connect(download_empty_signatures(feature))

            def download_unsigned_pv(parcel):
                def ret():
                    generated_pv_filename = os.path.join(
                        tempfile.gettempdir(),
                        "inventory_pv_non_signe_{}.pdf".format(parcel['id']))

                    pv_generator = config['pv_generator_cls'](
                        self.plugin_dir, parcel)
                    pv_generator.create_pdf_at(generated_pv_filename)

                    QDesktopServices.openUrl(
                        QUrl.fromLocalFile(generated_pv_filename))
                return ret

            self.upload_doc_dlg.download_unsigned_pv.clicked.connect(
                download_unsigned_pv(feature))

            self.upload_doc_dlg.exec_()
        else:
            msg_bar = self.iface.messageBar()
            msg_bar.pushCritical('Erreur', 'veuillez sélectionner un 1 et 1 \
seul élément')

        self.actions['pv_upload_signatures'].setChecked(False)

    def pv_w_signatures_print(self):
        config = self.paggf_actions_configuration['pv_w_signatures_print']
        parcel_layer = self.layers.get(
            config['pg_schema_name'], config['pg_table_name'])

        if parcel_layer.selectedFeatureCount() == 1:
            from .deps.PyPDF2__v1_26_0.PyPDF2 import PdfFileMerger
            merger = PdfFileMerger()
            parcel = parcel_layer.selectedFeatures()[0]

            if parcel['doc_idc_report_signatures']:
                generated_pv_filename = os.path.join(
                    tempfile.gettempdir(),
                    "inventory_generated_pv_{}.pdf".format(parcel['id']))

                signalture_filename, signature_b64data = \
                    parcel['doc_idc_report_signatures'].split(';')

                signature_filename = os.path.join(
                    tempfile.gettempdir(), signalture_filename)

                pv_w_signatues_filename = os.path.join(
                    tempfile.gettempdir(),
                    "inventory_pv_{}.pdf".format(parcel['id']))

                pv_generator = config['pv_generator_cls'](
                    self.plugin_dir, parcel)
                pv_generator.create_pdf_at(generated_pv_filename)

                with open(generated_pv_filename, "rb") as generated_pv_file,\
                        open(signature_filename, 'wb+') as signature_file,\
                        open(pv_w_signatues_filename, 'wb') as pv_w_signatures_file:
                    merger.append(fileobj=generated_pv_file, pages=(0, 4))
                    signature_file.write(
                        base64.b64decode(signature_b64data))
                    merger.append(signature_file)
                    merger.write(pv_w_signatures_file)

                QDesktopServices.openUrl(
                    QUrl.fromLocalFile(pv_w_signatues_filename))
            else:
                msg_bar = self.iface.messageBar()
                msg_bar.pushCritical(
                    'Erreur', "cette parcelle n'a pas de signatures, vous ne \
pouvez l'imprimer")
        else:
            msg_bar = self.iface.messageBar()
            msg_bar.pushCritical(
                'Erreur', 'veuillez sélectionner un 1 et 1 seul élément')

        self.actions['pv_w_signatures_print'].setChecked(False)

    def pan_trigger(self):
        self.iface.actionPan().trigger()

    def loadattributes_callback(self):
        if self.iface.activeLayer().selectedFeatureCount() == 1:
            self.load_attributes_dlg = LoadAttrFromFile(
                self.iface.activeLayer().selectedFeatures()[0],
                self.iface.activeLayer())
            self.load_attributes_dlg.exec_()
        else:
            msg_bar = self.iface.messageBar()
            msg_bar.pushCritical(
                'Erreur',
                'Veuillez sélectionner une et une seule parcelle')
