import os

from PyQt5 import uic
from PyQt5.QtWidgets import QDialog, QWidget, QTableWidgetItem, QHeaderView, \
     QFileDialog
from qgis.core import QgsExpressionContextUtils, QgsProject, QgsPoint, \
    QgsGeometry, QgsMessageLog, QgsVectorLayer, QgsCoordinateTransform, \
    QgsCoordinateReferenceSystem, QgsPoint, QgsGeometry
from PyQt5.QtGui import QBrush, QColor
from PyQt5.QtCore import QSettings

from ..db import get_connection
from ..paggf_layers import LoadedLayers
from .dlg_with_shp_loader import DlgWithShpLoader


FORM_CLASS_FILE_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'CID_load_parcel_file_widget.ui'))


class CIDLoadParcelFileWidget(QWidget, FORM_CLASS_FILE_W):
    def __init__(self, parent=None):
        super(CIDLoadParcelFileWidget, self).__init__(parent)
        self.setupUi(self)
        self.parent = parent

        self.shp_browse_btn.clicked.connect(self.shapefile_browse)

    def shapefile_browse(self):
        s = QSettings("PAGGF", "PAGGF")
        last_shp_path = s.value("lastSHPParcelPath", ".")

        if not os.path.exists(last_shp_path):
            last_shp_path = os.getcwd()
        shp_path, _ = QFileDialog.getOpenFileName(
            self, "Ouvrir le fichier SHP", last_shp_path, "*.shp *.csv *gpx")

        s.setValue("lastSHPParcelPath", shp_path)
        QgsMessageLog.logMessage(shp_path, 'PAGG', QgsMessageLog.ALL)

        self.shp_path_field.setText(shp_path)

    def shp_path(self):
        return self.shp_path_field.text()

    def get_values_for_db(self):
        layer_file_path = self.shp_path()
        layer_file_path_ext = os.path.splitext(layer_file_path)[1]

        provider_lib = "ogr"  # Default provider lib  for shp
        path = self.shp_path()  # Default uri for shp
        if layer_file_path_ext.lower() == '.csv':
            provider_lib = "delimitedtext"
            path = "\
file://{}?delimiter={}&crs=epsg:32735&xField={}&yField={}".format(
                self.shp_path(),
                ";", "X", "Y")
        elif layer_file_path_ext.lower() == '.gpx':
            provider_lib = "gpx"
            path = "{}?type=waypoint".format(self.shp_path())

        layer = QgsVectorLayer(
            path, "data_import", provider_lib)

        srid_layer = layer.crs().postgisSrid()

        transf = None
        if srid_layer != 32735:
            transf = QgsCoordinateTransform(
                QgsCoordinateReferenceSystem(srid_layer),
                QgsCoordinateReferenceSystem(32735))


        values = []
        for f in layer.getFeatures():
            if transf:
                geom = QgsGeometry(QgsPoint(transf.transform(f.geometry().asPoint())))
            else:
                geom = f.geometry()

            values.append((f['name'], geom))

        return values


FORM_CLASS_MANUAL_W, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'CID_load_parcel_manual_widget.ui'))


class CIDLoadParcelManualWidget(QWidget, FORM_CLASS_MANUAL_W):
    def __init__(self, parent=None):
        super(CIDLoadParcelManualWidget, self).__init__(parent)
        self.setupUi(self)
        self.parent = parent

        self.tableWidget.setRowCount(1)
        self.tableWidget.setColumnCount(3)

        self.tableWidget.setHorizontalHeaderItem(0, QTableWidgetItem('X'))
        self.tableWidget.setHorizontalHeaderItem(1, QTableWidgetItem('Y'))
        self.tableWidget.setHorizontalHeaderItem(2, QTableWidgetItem('Name'))

        self.tableWidget.horizontalHeader().setSectionResizeMode(
            QHeaderView.Stretch)

        self.tableWidget.setItem(0, 0, QTableWidgetItem())
        self.tableWidget.setItem(0, 1, QTableWidgetItem())
        self.tableWidget.setItem(0, 2, QTableWidgetItem())

        self.pushButton.clicked.connect(self.add_row)
        self.tableWidget.cellChanged.connect(self.check_cell)

    def check_cell(self, row, column):
        error = False
        for row in range(self.tableWidget.rowCount()):
            for column in range(2):
                item = self.tableWidget.item(row, column)
                if item:  # si possede widget
                    if item.text() and not item.text().isdigit():
                        item.setBackground(QBrush(QColor(200, 0, 0)))
                        error = True
                    else:
                        item.setBackground(QBrush())

        if self.parent:
            if error:
                self.parent.cancel_ok_btns.setEnabled(False)
            else:
                self.parent.cancel_ok_btns.setEnabled(True)

    def add_row(self):
        row_count = self.tableWidget.rowCount()

        self.tableWidget.setRowCount(self.tableWidget.rowCount() + 1)
        self.tableWidget.setItem(row_count, 0, QTableWidgetItem())
        self.tableWidget.setItem(row_count, 1, QTableWidgetItem())
        self.tableWidget.setItem(row_count, 2, QTableWidgetItem())

    def get_values_for_db(self):
        """
        retourne un tableau de couple (name, geom) à ajouter dans la db
        """
        values = []

        for row in range(self.tableWidget.rowCount()):
            x = self.tableWidget.item(row, 0).text()
            y = self.tableWidget.item(row, 1).text()
            if x and y:  # eviter cellule vides
                name = self.tableWidget.item(row, 2).text()
                geom = QgsGeometry(QgsPoint(float(x), float(y)))

                values.append((name, geom))
                # wkt = 'srid=32735;' + geom.exportToWkt()
                # values.append("('{}','{}'::geometry)".format(name, wkt))
        return values


FORM_CLASS, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'CID_load_parcel.ui'))


class CIDLoadParcelDlg(QDialog, FORM_CLASS):
    def __init__(
            self, mapCanvas, data_loader_class=CIDLoadParcelManualWidget, parent=None):
        """Constructor.

        This dialog has to know the id of municipality on which the user
        works. This id is stored in the variable 'municipality_id'.
        """
        self.mapCanvas = mapCanvas
        self.data_loader_class = data_loader_class

        super(CIDLoadParcelDlg, self).__init__(parent)
        self.setupUi(self)

        l = self.layout()
        self.data_loader = data_loader_class(parent=self)
        l.replaceWidget(self.widget, self.data_loader)

        self.project = QgsProject.instance()
        self.layers = LoadedLayers()
        self.municipality_id = QgsExpressionContextUtils \
            .projectScope(self.project).variable('id_municipality')
        self.municipality_name = None

        municipality_layer = self.layers.get("ref", "municipalities")
        if municipality_layer:
            for f in municipality_layer.getFeatures():
                if f['id'] == self.municipality_id:
                    self.municipality_name = f['name']

        if self.municipality_name:
            self.municipality_field.setText(self.municipality_name)
        else:
            self.municipality_field.setText('Inconnu')

        hill_layer = self.layers.get("ref", "hill_districts")
        if hill_layer:
            for f in hill_layer.getFeatures():
                if f['id'].startswith(self.municipality_id):
                    self.hill_field.addItem(
                        "{} - {}".format(f['id'], f['name']), f['id'])

        self.subhills = {}
        subhill_layer = self.layers.get("ref", "sub_hill_avenues")
        if subhill_layer:
            for f in subhill_layer.getFeatures():
                self.subhills[f['id']] = f['name']

        self.hill_field.currentIndexChanged[int].connect(
            self.when_hill_selection_changed)

        self.when_hill_selection_changed(0)

        self.cancel_ok_btns.accepted.connect(self.accepted_slot)
        self.cancel_ok_btns.rejected.connect(self.rejected_slot)

    def when_hill_selection_changed(self, index):
        """
        Called when the user change the selection of the hill (the selected
        hill in the combo-box hill_field has changed)

        Params
        ------
        int index : new index in the combobox of the selected hill
        """
        hill_id = self.hill_field.itemData(index)
        self.subhill_field.clear()
        for id_, name in self.subhills.items():
            if id_.startswith(hill_id):
                self.subhill_field.addItem("{} - {}".format(id_, name), id_)

        # mise à jour numéro parcelle
        conn = get_connection()
        cur = conn.cursor()
        cur.execute("select gis.next_parcel_num('{}')".format(hill_id))
        parcel_number = cur.fetchone()[0]
        self.parcel_number_field.setText(str(parcel_number))

    def subhill(self):
        return int(self.subhill_field.currentData())

    def parcel_number(self):
        return int(self.parcel_number_field.text())

    def accepted_slot(self):
        conn = get_connection()
        cur = conn.cursor()

        subhill = self.subhill()
        parcel_number = self.parcel_number()

        conn = get_connection()
        cur = conn.cursor()
        cur.execute(
            "\
insert into gis.parcels_to_valid (id_sub_hill_avenue, num_parcel) \
values (%s,%s) returning id",
            (subhill, parcel_number))
        id_parcel = cur.fetchone()[0]

        cur.execute("delete from gis.vertices_to_import")

        values = []
        for name, geom in self.data_loader.get_values_for_db():
            wkt = 'srid=32735;' + geom.exportToWkt()
            values.append("('{}','{}'::geometry)".format(name, wkt))

        query = "insert into gis.vertices_to_import (name, geom) values "
        query += ','.join(values)
        cur.execute(query)

        cur.execute("select gis.import_idc_vertices({})".format(id_parcel))
        conn.commit()

        l = self.layers.get("gis", "idc_parcels_to_valid")
        # update extent of affected layers
        if l:
            l.updateExtents()

            idc_vertices_to_valid = self.layers.get(
                "gis", "idc_vertices_to_valid")
            if idc_vertices_to_valid:
                idc_vertices_to_valid.updateExtents()

            isolated_vertices = self.layers.get("gis", "isolated_vertices")
            if isolated_vertices:
                isolated_vertices.updateExtents()

            segments_idc_parcels_to_valid = self.layers.get(
                "gis", "segments_idc_parcels_to_valid")
            if segments_idc_parcels_to_valid:
                segments_idc_parcels_to_valid.updateExtents()

            # zoom to the new feature
            self.mapCanvas.zoomToFeatureIds(l, [id_parcel])



    def rejected_slot(self):
        self.close()
